#include <PubSubClient.h>
#include <ESP8266WiFi.h>
#include <FastLED.h>
#include <ESP8266HTTPClient.h>
#include <ArduinoJson.h>
#include <time.h>

#include "sensors.h"

#define HOSTNAME "Fishtank"
#define MQTT_SERVER "192.168.1.3"
#define SSID "GCHQ Surveillance Van"
#define PASSWORD "cocklol."
#define LIGHT_PIN D2
#define LED_COUNT 36
#define PH_TOPIC "/home/sensors/fishtank/ph"
#define TEMPERATURE_TOPIC "/home/sensors/fishtank/temperature"
#define TEMPERATURE_PIN D5

#define SOLAR_EVENT_URL "http://api.sunrise-sunset.org/json?lat=20.548103&lng=96.916835&formatted=0"
#define TIMEZONE_OFFSET 27000 // 7.5 hours in seconds
#define NTP_POOL "uk.pool.ntp.org"
#define HEAT_INDEX_MAX 241.0 // After 241 the pallet seems to go dim

void mqttCallback(char *topic, byte *payload, unsigned int length);

WiFiClient _wifiClient;
PubSubClient _mqttClient = PubSubClient(MQTT_SERVER, 1883, mqttCallback, _wifiClient);

Sensors _sensors = Sensors(TEMPERATURE_PIN, TEMPERATURE_TOPIC, PH_TOPIC, _mqttClient);

CRGB _leds[LED_COUNT];
int _lightsOnStartTime, _lightsOnEndTime, _lightsOffStartTime, _lightsOffEndTime;


void waitForWiFi() {
	if (WiFi.status() == WL_CONNECTED)
		return;
	Serial.println("Reconnecting WiFi...");
	int waitCount=0;
	while (WiFi.status() != WL_CONNECTED) {
		delay(100);
		waitCount++;
		if (waitCount>600)
			ESP.restart();
	}
	Serial.println("Connected WiFi!");
}

void waitForMQTT() {
	if (_mqttClient.connected())
		return;
	Serial.println("Reconnecting MQTT...");
	while (!_mqttClient.connected())
		if (!_mqttClient.connect(HOSTNAME))
			delay(50);
	Serial.println("Connected MQTT!");
}

void reconnect() {
	waitForWiFi();
	waitForMQTT();
}

String getSolarEventJson(){
	waitForWiFi();

	HTTPClient http;
	http.begin(SOLAR_EVENT_URL);
	int httpCode = http.GET();
	String result = "";
	if (httpCode > 0)
		result = http.getString();
	http.end();

	return result;
}

int getEpoch(const char *str) {
	struct tm time;
	int year, month;
	sscanf(str, "%d-%d-%dT%d:%d:%d+00:00", &year, &month, &time.tm_mday, &time.tm_hour, &time.tm_min, &time.tm_sec);
	time.tm_year = year - 1900;
	time.tm_mon = month - 1;
	return mktime(&time);
}

int adjustForTimezone(int epoch){
	return epoch + TIMEZONE_OFFSET;
}

void parseEvents(){
	String json = getSolarEventJson();
	StaticJsonBuffer<1024> jsonBuffer;
	JsonObject &root = jsonBuffer.parseObject(json);

	if (!root.success()) {
		Serial.println("Could not parse json");
		return;
	}

	const char *astronomicalTwilightBegin = root["results"]["astronomical_twilight_begin"];
	const char *sunrise = root["results"]["sunrise"];
	const char *sunset = root["results"]["sunset"];
	const char *astronomicalTwilightEnd = root["results"]["astronomical_twilight_end"];

	_lightsOnStartTime = adjustForTimezone(getEpoch(astronomicalTwilightBegin));
	_lightsOnEndTime = adjustForTimezone(getEpoch(sunrise));
	_lightsOffStartTime = adjustForTimezone(getEpoch(sunset));
	_lightsOffEndTime = adjustForTimezone(getEpoch(astronomicalTwilightEnd));
}

void setup() {
	Serial.begin(115200);

	_sensors.setup();

	WiFi.hostname(HOSTNAME);
	WiFi.begin(SSID, PASSWORD);
	reconnect();

	delay(2000);
	configTime(0, 0, NTP_POOL);

	parseEvents();

	FastLED.addLeds<WS2811, LIGHT_PIN, GRB>(_leds, LED_COUNT).setCorrection(TypicalSMD5050).setTemperature(Tungsten40W);
}

float getIndexMultiplier(int now, int startTime, int endTime, bool swap){
	float indexMultiplier = 1;
	float a = swap ? endTime - now : now - startTime;
	float b = endTime - startTime;
	return constrain(a / b, 0, 1);
}

float getHeatIndex() {
	time_t now = time(nullptr);
	float indexMultiplier = now >= _lightsOffStartTime ? getIndexMultiplier(now, _lightsOffStartTime, _lightsOffEndTime, true) : getIndexMultiplier(now, _lightsOnStartTime, _lightsOnEndTime, false);
	return HEAT_INDEX_MAX * indexMultiplier;
}

void doLighting() {
	EVERY_N_SECONDS(1) {
		CRGB colour = ColorFromPalette(HeatColors_p, getHeatIndex());
		fill_solid(_leds, LED_COUNT, colour);
		FastLED.show();
	}

	EVERY_N_SECONDS(10800){
		parseEvents();
	}
}

void publishReadings(){
	EVERY_N_SECONDS(60) {
		_sensors.publishpH();
		_sensors.publishTemperature();
	}
}

void syncTime(){
	EVERY_N_SECONDS(3600) {
		configTime(0, 0, NTP_POOL);
	}
}

void loop() {
	reconnect();
	_mqttClient.loop();

	syncTime();

	doLighting();

	//publishReadings();

	delay(10);
}

void mqttCallback(char *topic, byte *payload, unsigned int length) {
	Serial.print("Got payload: ");
	for (int i = 0; i < length; i++)
	{
		Serial.print((char)payload[i]);
	}
	Serial.println();
}